#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import platform
import os.path as osp
from setuptools import setup, find_namespace_packages

# Ensure user has the correct Python version
if sys.version_info < (3, 8):
    print("Mathics support Python 3.8 and above; you have %d.%d" % sys.version_info[:2])
    sys.exit(-1)


def get_srcdir():
    filename = osp.normcase(osp.dirname(osp.abspath(__file__)))
    return osp.realpath(filename)


def read(*rnames):
    return open(osp.join(get_srcdir(), *rnames)).read()


# stores __version__ in the current namespace
exec(compile(open("pymathics/graph/version.py").read(), "version.py", "exec"))
long_description = read("README.rst") + "\n"

is_PyPy = platform.python_implementation() == "PyPy"

setup(
    name="pymathics-graph",
    version=__version__,  # noqa
    packages=find_namespace_packages(include=["pymathics.*"]),
    install_requires=[
        "Mathics3>=7.0.0.dev0",
        "networkx>=3.0.0",
        "pydot",
        "matplotlib",
        "scipy>=1.10.0",
    ],
    # don't pack Mathics in egg because of media files, etc.
    zip_safe=False,
    maintainer="Mathics3 Group",
    maintainer_email="rb@dustyfeet.com",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # metadata for upload to PyPI
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Interpreters",
    ],
    # TODO: could also include long_description, download_url,
)



"""
Unit tests for pymathics.graph.measures_and_metrics
"""

from test.helper import check_evaluation


def test_graph_distance():
    for str_expr, str_expected, mess in [
        (
            "GraphDistance[{1 <-> 2, 2 <-> 3, 3 <-> 4, 2 <-> 4, 4 -> 5}, 1, 5]",
            "3",
            None,
        ),
        ("GraphDistance[{1 <-> 2, 2 <-> 3, 3 <-> 4, 4 -> 2, 4 -> 5}, 1, 5]", "4", None),
        (
            "GraphDistance[{1 <-> 2, 2 <-> 3, 4 -> 3, 4 -> 2, 4 -> 5}, 1, 5]",
            "Infinity",
            None,
        ),
        (
            "Sort[GraphDistance[{1 <-> 2, 2 <-> 3, 3 <-> 4, 2 <-> 4, 4 -> 5}, 3]]",
            "{0, 1, 1, 2, 2}",
            None,
        ),
        (
            "GraphDistance[{}, 1, 1]",
            "GraphDistance[{}, 1, 1]",
            [
                "The vertex at position 2 in GraphDistance[{}, 1, 1] does not belong to "
                "the graph at position 1."
            ],
        ),
        (
            "GraphDistance[{1 -> 2}, 3, 4]",
            "GraphDistance[{1 -> 2}, 3, 4]",
            [
                "The vertex at position 2 in GraphDistance[{1 -> 2}, 3, 4] does not belong "
                "to the graph at position 1."
            ],
        ),
    ]:
        check_evaluation(str_expr, str_expected, expected_messages=mess)



# -*- coding: utf-8 -*-

"""
Collection classes
"""


from pymathics.graph.graphsymbols import SymbolDirectedEdge

# There is no user-facing documentation here.
no_doc = True


def _count_edges(counts, edges, sign):
    n_directed, n_undirected = counts
    for edge in edges:
        if edge.head is SymbolDirectedEdge:
            n_directed += sign
        else:
            n_undirected += sign
    return n_directed, n_undirected


class _Collection:
    def __init__(self, expressions, properties=None, index=None):
        self.expressions = expressions
        self.properties = properties if properties else None
        self.index = index

    def clone(self):
        properties = self.properties
        return _Collection(
            self.expressions[:], properties[:] if properties else None, None
        )

    def filter(self, expressions):
        index = self.get_index()
        return [expr for expr in expressions if expr in index]

    def extend(self, expressions, properties):
        if properties:
            if self.properties is None:
                self.properties = [None] * len(self.expressions)
            self.properties.extend(properties)
        self.expressions.extend(expressions)
        self.index = None
        return expressions

    def delete(self, expressions):
        index = self.get_index()
        trash = set(index[x] for x in expressions)
        deleted = [self.expressions[i] for i in trash]
        self.expressions = [x for i, x in enumerate(self.expressions) if i not in trash]
        self.properties = [x for i, x in enumerate(self.properties) if i not in trash]
        self.index = None
        return deleted

    def data(self):
        return self.expressions, list(self.get_properties())

    def get_index(self):
        index = self.index
        if index is None:
            index = dict((v, i) for i, v in enumerate(self.expressions))
            self.index = index
        return index

    def get_properties(self):
        if self.properties:
            for p in self.properties:
                yield p
        else:
            for _ in range(len(self.expressions)):
                yield None

    def get_sorted(self):
        index = self.get_index()
        return lambda c: sorted(c, key=lambda v: index[v])

    def get_property(self, element, name):
        properties = self.properties
        if properties is None:
            return None
        index = self.get_index()
        i = index.get(element)
        if i is None:
            return None
        p = properties[i]
        if p is None:
            return None
        return p.get(name)


class _EdgeCollection(_Collection):
    def __init__(
        self, expressions, properties=None, index=None, n_directed=0, n_undirected=0
    ):
        super(_EdgeCollection, self).__init__(expressions, properties, index)
        self.counts = (n_directed, n_undirected)

    def is_mixed(self):
        n_directed, n_undirected = self.counts
        return n_directed > 0 and n_undirected > 0

    def clone(self):
        properties = self.properties
        n_directed, n_undirected = self.counts
        return _EdgeCollection(
            self.expressions[:],
            properties[:] if properties else None,
            None,  # index
            n_directed,
            n_undirected,
        )

    def extend(self, expressions, properties):
        added = super(_EdgeCollection, self).extend(expressions, properties)
        self.counts = _count_edges(self.counts, added, 1)
        return added

    def delete(self, expressions):
        deleted = super(_EdgeCollection, self).delete(expressions)
        self.counts = _count_edges(self.counts, deleted, -1)
        return deleted



"""
Graph Measures and Metrics

Measures include basic measures, such as the number of vertices and edges, \
connectivity, degree measures, centrality, and so on.
"""

from typing import Optional

import networkx as nx
from mathics.core.atoms import Integer, Integer1, Integer2, Integer3
from mathics.core.convert.expression import ListExpression, to_mathics_list
from mathics.core.evaluation import Evaluation
from mathics.core.expression import Expression, from_python
from mathics.core.systemsymbols import SymbolCases, SymbolDirectedInfinity, SymbolLength

from pymathics.graph.base import _NetworkXBuiltin

# FIXME put this in its own file/module basic
# when pymathics doc can handle this.
# """
# Basic Graph Measures
# """


class _PatternCount(_NetworkXBuiltin):
    """
    Counts of vertices or edges, allowing rules to specify the graph.
    """

    no_doc = True

    def eval(self, graph, expression, evaluation, options) -> Optional[Integer]:
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return Integer(len(self._items(graph)))

    def eval_patt(
        self, graph, patt, expression, evaluation, options
    ) -> Optional[Expression]:
        "%(name)s[graph_, patt_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return Expression(
                SymbolLength,
                Expression(
                    SymbolCases,
                    ListExpression(*(from_python(item) for item in self._items(graph))),
                    patt,
                ),
            )


class EdgeCount(_PatternCount):
    """
    <url>
    :NetworkX:
    https://networkx.org/documentation/latest/reference/generated/networkx.classes.function.edges.html#edges</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/EdgeCount.html</url>

    <dl>
       <dt>'EdgeCount[$g$]'
       <dd>returns a count of the number of edges in graph $g$.

       <dt>'EdgeCount[$g$, $patt$]'
       <dd>returns the number of edges that match the pattern $patt$.

       <dt>'EdgeCount[{$v$->$w}, ...}, ...]'
       <dd>uses rules $v$->$w$ to specify the graph $g$.
    </dl>

    >> EdgeCount[{1 -> 2, 2 -> 3}]
     = 2
    """

    no_doc = False
    summary_text = "count edges in graph"

    def _items(self, graph):
        return graph.G.edges


class GraphDistance(_NetworkXBuiltin):
    """
    <url>
    :NetworkX:
    https://networkx.org/documentation/latest/reference/algorithms/generated/networkx.algorithms.shortest_paths.generic.shortest_path_length.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/GraphDistance.html</url>

    <dl>
      <dt>'GraphDistance'[$g$, $s$, $t$]
      <dd>returns the distance from source vertex $s$ to target vertex $t$ in the graph $g$.
    </dl>

    <dl>
      <dt>'GraphDistance[$g$, $s$]'
      <dd>returns the distance from source vertex $s$ to all vertices in the graph $g$.
    </dl>

    <dl>
      <dt>'GraphDistance[{$v$->$w$, ...}, ...]'
      <dd>use rules $v$->$w$ to specify the graph $g$.
    </dl>

    >> g = Graph[{1 -> 2, 2 <-> 3, 4 -> 3, 2 <-> 4, 4 -> 5}, VertexLabels->True]
     = -Graph-

    >> GraphDistance[g, 1, 5]
     = 3

    >> GraphDistance[g, 4, 2]
     = 1

    >> GraphDistance[g, 5, 4]
     = Infinity

    >> GraphDistance[g, 5]
     = {Infinity, Infinity, Infinity, Infinity, 0}

    >> GraphDistance[g, 3]
     = {Infinity, 1, 2, 0, 3}

    >> GraphDistance[g, 4]
     = {Infinity, 1, 0, 1, 1}
    """

    summary_text = "get path distance"

    def eval_s(
        self, graph, s, expression, evaluation: Evaluation, options: dict
    ) -> Optional[ListExpression]:
        "GraphDistance[graph_, s_, OptionsPattern[GraphDistance]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            weight = graph.update_weights(evaluation)
            d = nx.shortest_path_length(graph.G, source=s, weight=weight)
            inf = Expression(SymbolDirectedInfinity, Integer1)
            return to_mathics_list(*[d.get(v, inf) for v in graph.vertices])

    def eval_s_t(self, graph, s, t, expression, evaluation: Evaluation, options: dict):
        "GraphDistance[graph_, s_, t_, OptionsPattern[GraphDistance]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if not graph:
            return
        G = graph.G
        if not G.has_node(s):
            self._not_a_vertex(expression, Integer2, evaluation)
        elif not G.has_node(t):
            self._not_a_vertex(expression, Integer3, evaluation)
        else:
            try:
                weight = graph.update_weights(evaluation)
                return from_python(
                    nx.shortest_path_length(graph.G, source=s, target=t, weight=weight)
                )
            except nx.exception.NetworkXNoPath:
                return Expression(SymbolDirectedInfinity, Integer1)


class VertexCount(_PatternCount):
    """
    <url>
    :NetworkX:
    https://networkx.org/documentation/latest/reference/generated/networkx.classes.function.nodes.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/VertexCount.html</url>

    <dl>
       <dt>'VertexCount[$g$]'
       <dd>returns a count of the number of vertices in graph $g$.

       <dt>'VertexCount[$g$, $patt$]'
       <dd>returns the number of vertices that match the pattern $patt$.

       <dt>'VertexCount[{$v$->$w}, ...}, ...]'
       <dd>uses rules $v$->$w$ to specify the graph $g$.
    </dl>

    >> VertexCount[{1 -> 2, 2 -> 3}]
     = 3

    >> VertexCount[{1 -> x, x -> 3}, _Integer]
     = 2
    """

    no_doc = False
    summary_text = "count vertices in graph"

    def _items(self, graph):
        return graph.G.nodes


# Put this in its own file/module "degree.py"
# when pymathics doc can handle.
# """
# Graph Degree Measures
# """


class VertexDegree(_NetworkXBuiltin):
    """
    <url>
    :NetworkX:
    https://networkx.org/documentation/latest/reference/classes/generated/networkx.Graph.degree.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/VertexDegree.html</url>

    <dl>
       <dt>'VertexDegree[$g$]'
       <dd>returns a list of the degrees of each of the vertices in graph $g$.

       <dt>'EdgeCount[$g$, $patt$]'
       <dd>returns the number of edges that match the pattern $patt$.

       <dt>'EdgeCount[{$v$->$w}, ...}, ...]'
       <dd>uses rules $v$->$w$ to specify the graph $g$.
    </dl>

    >> VertexDegree[{1 <-> 2, 2 <-> 3, 2 <-> 4}]
     = {1, 3, 1, 1}
    """

    no_doc = False
    summary_text = "list graph vertex degrees"

    def eval(self, graph, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"

        def degrees(graph):
            degrees = dict(list(graph.G.degree(graph.vertices)))
            return ListExpression(*[Integer(degrees.get(v, 0)) for v in graph.vertices])

        return self._evaluate_atom(graph, options, degrees)


# TODO: VertexInDegree, VertexOutDegree
